'use strict';

function _defineProperty(obj, key, value) { if (key in obj) { Object.defineProperty(obj, key, { value: value, enumerable: true, configurable: true, writable: true }); } else { obj[key] = value; } return obj; }

var format = require('format-io');
var exec = require('execon');
var squad = require('squad');
var shortdate = require('shortdate');
var promisify = require('es6-promisify');
var currify = require('currify/legacy');
var zames = require('zames/legacy');

var WIN = process.platform === 'win32';
var BROWSER = typeof window !== 'undefined';

var sort = currify(function (fn, array) {
    return array.sort(fn);
});
var getStat = currify(_getStat);
var parseAllStats = currify(_parseAllStats);
var replaceProperty = currify(_replaceProperty);

var getAllStats = zames(_getAllStats);

var getFS = function getFS() {
    if (!BROWSER) return require('fs');

    var Filer = require('filer');
    return new Filer.FileSystem();
};

var fs = getFS();
var nicki = !WIN && !BROWSER && require('nicki/legacy');

var readdir = promisify(fs.readdir, fs);

/* sorting on Win and node v0.8.0 */
var sortFiles = sort(function (a, b) {
    return a.name > b.name ? 1 : -1;
});

var good = function good(f) {
    return function () {
        for (var _len = arguments.length, a = Array(_len), _key = 0; _key < _len; _key++) {
            a[_key] = arguments[_key];
        }

        return f.apply(undefined, [null].concat(a));
    };
};

module.exports = readify;

function readify(path, type, fn) {
    if (!fn) {
        fn = type;
        type = '';
    }

    check(path, fn);

    readdir(path).then(getAllStats(path, type)).then(good(fn)).catch(fn);
}

function check(path, callback) {
    var pathMsg = 'path should be string!';
    var callbackMsg = 'callback should be function!';

    if (typeof path !== 'string') throw Error(pathMsg);

    if (typeof callback !== 'function') throw Error(callbackMsg);
}

/**
 * @param path
 * @param names
 */
function _getAllStats(path, type, names, callback) {
    var length = names.length;
    var dir = format.addSlashToEnd(path);

    if (!length) return fillJSON(dir, [], type, callback);

    var funcs = names.map(function (name) {
        return getStat(name, dir + name);
    });

    exec.parallel(funcs, function () {
        for (var _len2 = arguments.length, args = Array(_len2), _key2 = 0; _key2 < _len2; _key2++) {
            args[_key2] = arguments[_key2];
        }

        var files = args.slice(1);
        fillJSON(dir, files, type, callback);
    });
}

function emptyStat() {
    return {
        mode: 0,
        size: 0,
        mtime: 0,
        uid: 0,
        isDirectory: function isDirectory() {}
    };
}

function _getStat(name, path, callback) {
    fs.stat(path, function (error) {
        var data = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : emptyStat();

        data.name = name;

        callback(null, data);
    });
}

function replaceDate(stat) {
    var date = !stat.date ? '' : shortdate(stat.date, {
        order: 'little'
    });

    return Object.assign(stat, {
        date: date
    });
};

function replaceMode(stat) {
    var octal = Number(stat.mode).toString(8);
    var mode = format.permissions.symbolic(octal);

    return Object.assign(stat, {
        mode: mode
    });
}

function replaceSize(stat) {
    var size = format.size(stat.size);

    return Object.assign(stat, {
        size: size
    });
}

function _parseAllStats(type, array) {
    var parse = function parse(item) {
        return parseStat(type, item);
    };

    var files = array.map(parse);

    if (type === 'raw') return files;

    return files.map(replaceDate).map(replaceSize).map(replaceMode);
}

function parseStat(type, stat) {
    var isDir = stat.isDirectory();
    var size = isDir ? 'dir' : stat.size;

    return {
        name: stat.name,
        size: size,
        date: stat.mtime,
        owner: stat.uid,
        mode: stat.mode
    };
}

/**
 * Function fill JSON by file stats
 *
 * @param params - { files, stats, path }
 */
function fillJSON(path, stats, type, callback) {
    var processFiles = squad(changeOrder, sortFiles, parseAllStats(type));
    var json = {
        path: '',
        files: processFiles(stats)
    };

    json.path = format.addSlashToEnd(path);

    if (type === 'raw') return callback(null, json);

    changeUIDToName(json, function (error, files) {
        json.files = files || json.files;
        callback(null, json);
    });
}

function changeUIDToName(json, callback) {
    if (!nicki) return callback(null, json.files);

    nicki(function (error, names) {
        if (error) return callback(error);

        var replaceOwner = replaceProperty(names, 'owner');
        var files = json.files.map(replaceOwner);

        callback(null, files);
    });
}

function isDir(file) {
    return file.size === 'dir';
}

function not(fn) {
    return function () {
        return !fn.apply(undefined, arguments);
    };
}

function changeOrder(json) {
    var isFile = not(isDir);
    var dirs = json.filter(isDir);
    var files = json.filter(isFile);
    var sorted = dirs.concat(files);

    return sorted;
}

function _replaceProperty(obj, prop, item) {
    var n = item[prop];
    var data = obj[n];

    if (typeof data === 'undefined') return item;

    return Object.assign(item, _defineProperty({}, prop, data));
}